import logging
import mimetypes
import os
from pathlib import Path
from typing import Union
from uuid import UUID

import requests
from requests import Session

from picsellia import exceptions
from picsellia.services.upload.adapters import (
    AzureUploadAdapter,
    GoogleMultipartUploadAdapter,
    S3MultipartUploadAdapter,
    S3UploadAdapter,
    UploadAdapter,
)
from picsellia.services.upload.platform import PlatformAdapter

logger = logging.getLogger("picsellia")

LARGE_FILE_SIZE = 25 * 1024 * 1024
DEFAULT_TIMEOUT = 30


class FileUploader:
    def __init__(self, connector_id: UUID, session: Session, host: str, headers: dict):
        self.platform_adapter = PlatformAdapter(connector_id, session, host, headers)

    def upload(
        self, object_name: str, path: Union[str, Path]
    ) -> tuple[requests.Response, bool, str]:
        is_large, content_type = self._read_file(path)

        if is_large:
            upload = self._upload_large_file(object_name, path, content_type)
        else:
            upload = self._upload_small_file(object_name, path, content_type)

        return upload, is_large, content_type

    def _upload_large_file(
        self,
        object_name: str,
        path: Union[str, Path],
        content_type: str,
    ):
        try:
            response = self.platform_adapter.init_multipart_upload(
                object_name, content_type
            ).json()
            adapter = self._build_multipart_adapter(response)
        except exceptions.BadRequestError:
            logger.warning(
                "This file is large but it is impossible to use multipart for this upload."
                "Trying to upload in only one upload."
            )
            return self._upload_small_file(object_name, path, content_type)

        return adapter.upload(object_name, path, content_type)

    def _upload_small_file(
        self, object_name: str, path: Union[str, Path], content_type: str
    ):
        response = self.platform_adapter.generate_presigned_url(
            object_name, content_type
        ).json()
        adapter = self._build_adapter(response)
        return adapter.upload(object_name, path, content_type)

    @staticmethod
    def _read_file(path: Union[str, Path]) -> tuple[bool, str]:
        if not os.path.isfile(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        is_large = Path(path).stat().st_size > LARGE_FILE_SIZE
        content_type = mimetypes.guess_type(path, strict=False)[0]
        if content_type is None:
            content_type = "application/octet-stream"

        return is_large, content_type

    def _build_adapter(self, response: dict) -> UploadAdapter:
        if (
            "client_type" not in response
            or "presigned_url_data" not in response
            or "url" not in response["presigned_url_data"]
            or "fields" not in response["presigned_url_data"]
        ):
            raise exceptions.DistantStorageError(
                "Platform could not generate a presigned url to upload this file."
            )

        client_type = response["client_type"].lower()
        url = response["presigned_url_data"]["url"]
        fields = response["presigned_url_data"]["fields"]
        if client_type in ["aws", "minio", "google"]:
            return S3UploadAdapter(self.platform_adapter, url, fields)
        elif client_type == "azure":
            return AzureUploadAdapter(self.platform_adapter, url, fields)
        else:
            raise exceptions.BadRequestError(
                f"This version of SDK cannot upload on {client_type} connector."
            )

    def _build_multipart_adapter(self, response: dict) -> UploadAdapter:
        if "client_type" not in response:
            raise exceptions.BadRequestError(
                "Platform cannot accept a multipart upload for this file."
            )

        client_type = response["client_type"].lower()
        if client_type in ["aws", "minio"]:
            upload_id = response["upload_id"]
            return S3MultipartUploadAdapter(self.platform_adapter, upload_id)
        elif client_type == "google":
            url = response["url"]
            return GoogleMultipartUploadAdapter(self.platform_adapter, url)
        else:
            raise exceptions.BadRequestError(
                f"This version of SDK cannot upload on {client_type} connector."
            )



import logging
import os
from pathlib import Path
from typing import Optional, Union
from uuid import UUID

from orjson import orjson
from PIL import Image

from picsellia.exceptions import UploadError
from picsellia.exif import read_image_metadata
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.datasource import DataSource
from picsellia.services.error_manager import ErrorManager
from picsellia.types.enums import DataType, ObjectDataType
from picsellia.utils import filter_payload

ALLOWED_IMAGE_CONTENT_TYPES = (
    "image/jpeg",
    "image/bmp",
    "image/gif",
    "image/png",
    "image/tiff",
)

logger = logging.getLogger("picsellia")


class DataUploader:
    def __init__(
        self,
        connexion: Connexion,
        datalake_id: UUID,
        connector_id: Optional[UUID],
        computed_tag_ids: list[UUID],
        source: Optional[DataSource],
        fill_metadata: bool,
        upload_dir: Optional[str],
        error_manager: Optional[ErrorManager],
    ):
        self.connexion = connexion
        self.connector_id = connector_id
        self.datalake_id = datalake_id
        self.error_manager = error_manager
        self.fill_metadata = fill_metadata
        self.computed_tags_ids = computed_tag_ids
        self.source = source
        self.upload_dir = upload_dir

    def upload(self, items: tuple[Union[str, Path], Optional[dict], Optional[dict]]):
        path, metadatum, custom_metadatum = items
        try:
            filename = os.path.basename(path)
            object_name = self.connexion.generate_data_object_name(
                filename, ObjectDataType.DATA, self.connector_id, self.upload_dir
            )
            _, _, content_type = self.connexion.upload_file(
                object_name, path, connector_id=self.connector_id
            )
            payload = self._prepare_payload(
                filename, object_name, path, content_type, metadatum, custom_metadatum
            )
            return self.connexion.post(
                f"/api/datalake/{self.datalake_id}/datas",
                data=orjson.dumps(payload),
            )
        except Exception as e:
            message = f"Could not upload path '{path}'"
            logger.error(message, exc_info=e)
            if self.error_manager:
                self.error_manager.append(
                    UploadError(message=message, path=path, parent=e)
                )
            return None

    def _prepare_payload(
        self,
        filename: str,
        object_name: str,
        path: str,
        content_type: str,
        metadatum: dict,
        custom_metadatum: Optional[dict],
    ):
        meta = None
        metadata = filter_payload(metadatum) if metadatum else None

        if content_type in ALLOWED_IMAGE_CONTENT_TYPES:
            data_type = DataType.IMAGE
            exif_data = self.prepare_metadata_from_image(
                path, metadatum, self.fill_metadata
            )
            if exif_data:
                meta = {"width": exif_data.width, "height": exif_data.height}
                metadata = exif_data.metadata
        elif content_type.startswith("video/"):
            data_type = DataType.VIDEO
        else:
            data_type = DataType.IMAGE
            logger.info(
                f"SDK cannot open {path} because its content type {content_type} is not openable here. "
                f"We will still upload your data and try handle it in our backend services."
            )

        payload = {
            "type": data_type,
            "filename": filename,
            "object_name": object_name,
            "content_type": content_type,
            "tags": self.computed_tags_ids,
            "data_source_id": self.source.id if self.source else None,
        }
        if meta:
            payload["meta"] = meta
        if metadata:
            payload["metadata"] = metadata
        if custom_metadatum:
            payload["custom_metadata"] = custom_metadatum
        return payload

    @staticmethod
    def prepare_metadata_from_image(path: str, metadatum: dict, fill_metadata: bool):
        try:
            # Try opening with Pillow to read exif metadata, width and height
            with Image.open(path) as image:
                return read_image_metadata(image, metadatum, fill_metadata)
        except Exception:
            logger.error(
                f"{path} could not be opened with Pillow, Picsellia will try to compute size of this image later."
            )
            return None
